from typing import Set, List, Dict
import torch
from torch.cuda import amp
from fastai.basics import store_attr, noop
import dill
from .utils import find_incremental_filename
from transformers import AdamW
from transformers.tokenization_utils_base import BatchEncoding
from .callbacks import ProgressBar


"""
Order of lifecycle:

- before_fit
    - before_epoch
            - before_train_batch
                - before_train_loss
                - after_train_loss
            - after_train_batch
        - before_validate (interval)
            - before_valid_batch
                - before_valid_loss
                - after_valid_loss
            - after_valid_batch
        - after_validate (interval)
    - after_epoch
- after_fit
"""


class Learner:
    def __init__(self, model, dls, config, valid_interval=None, cbs=None, get_pred=None, save_learner=True):
        store_attr(
            'model, dls, config, valid_interval, cbs, get_pred, save_learner', self)
        self.lr = config.lr
        self.optimizer = self.get_optimizer(config.optimizer)
        self.loss_func = self.get_loss_func(config.loss_func)

        if valid_interval is None:
            self.valid_interval = len(dls.train_dl)

        # save dls
        with open(f"{self.config.path}/dls.pkl", 'wb') as dls_file:
            dill.dump(dls, dls_file)
            print("saved dls successfully!")

        if cbs is None:
            cbs = []
        self.cbs = [ProgressBar()] + self.cbs
        for cb in self.cbs:
            cb.learner = self
        self.cb_dict = {type(cb).__name__: cb for cb in self.cbs}

    def get_loss_func(self, loss_func):
        if loss_func == 'cross_entropy_loss':
            return torch.nn.functional.cross_entropy
        elif loss_func == 'binary_cross_entropy_with_logits':
            return torch.nn.functional.binary_cross_entropy_with_logits
        elif loss_func == 'mse':
            return torch.nn.MSELoss()
        elif loss_func == 'bce':
            return torch.nn.BCELoss()
        else:
            raise ValueError(f"loss_func {loss_func} not supported")

    def get_optimizer(self, optimizer: str):
        if optimizer == 'adamw':
            return AdamW(self.model.parameters(), lr=self.lr)
        elif optimizer == 'adam':
            return torch.optim.Adam(self.model.parameters(), lr=self.lr)
        elif optimizer == 'sgd':
            return torch.optim.SGD(self.model.parameters(), lr=self.lr)
        else:
            raise ValueError(f"optimizer {optimizer} not supported")

    def get_batch_x_y(self, batch):
        # support huggingface dict
        if isinstance(batch, dict):
            batch_y = batch['label']
            del batch['label']
            batch_x = batch
        elif isinstance(batch, BatchEncoding):
            batch_y = batch['labels']
            del batch['labels']
            batch_x = batch
        else:
            batch_x, batch_y = batch
        return batch_x, batch_y

    def fit(self, epochs):
        self.epochs = epochs
        self('before_fit')
        for epoch in range(1, epochs+1):

            self.epoch = epoch
            self('before_epoch')
            self.model.train()
            for i, batch in enumerate(self.dls.train_dl):
                self.batch_x, self.batch_y = self.get_batch_x_y(batch)
                self.train_batch()
                if i % self.valid_interval == 0:
                    self.validate_interval()

            self('after_epoch')
        self('after_fit')

    def validate_interval(self):
        self('before_validate')
        self.model.eval()

        for batch in self.dls.valid_dl:
            self.batch_x, self.batch_y = self.get_batch_x_y(batch)
            self.validate_batch()

        self('after_validate')
        self.model.train()

    def train_batch(self):
        self('before_train_batch')
        # forward
        if self.get_pred == None:
            self.pred = self.model(self.batch_x)
        else:
            self.out = self.model(self.batch_x)
            self.pred = self.get_pred(self.out)
        self('before_train_loss')
        self.loss = self.loss_func(self.pred, self.batch_y)
        self('after_train_loss')
        # backward
        self.loss.backward()
        self.optimizer.step()
        self.optimizer.zero_grad()
        self('after_train_batch')

    def validate_batch(self):
        self('before_valid_batch')
        # forward
        with torch.no_grad():
            if self.get_pred == None:
                self.pred = self.model(self.batch_x)
            else:
                self.out = self.model(self.batch_x)
                self.pred = self.get_pred(self.out)
            self('before_valid_loss')
            self.loss = self.loss_func(self.pred, self.batch_y)
            self('after_valid_loss')
        self('after_valid_batch')

    def __call__(self, name):
        for cb in self.cbs:
            getattr(cb, name, noop)()

    def save(self):
        """
        Save callback dict to `path`
        """
        self.export = {}
        for cb_name, cb in self.cb_dict.items():
            # export everything except for reference to learner
            self.export[cb_name] = {key: value for key,
                                    value in cb.__dict__.items() if key != 'learner'}
        # save
        with open(f"{self.config.path}/learner.pkl", 'wb') as learner_file:
            dill.dump(self.export, learner_file)
            print("save successful!")


# TODO: update MixedPrecisionCallback to match the updated Learner above
class MixedPrecisionLearner(Learner):
    def fit(self, epochs):
        self.epochs = epochs
        self.scaler = amp.GradScaler()

        self('before_fit')
        for epoch in range(1, epochs+1):
            self.epoch = epoch
            self('before_epoch')
            self.model.train()
            for i, batch in enumerate(self.dls.train_dl):
                self.batch_x, self.batch_y = batch
                self.train_batch()
                if i % self.valid_interval == 0:
                    self.validate_interval()
            self('after_epoch')
        self('after_fit')

    def train_batch(self):
        self('before_train_batch')
        # forward
        with amp.autocast():
            self.pred = self.get_pred(self.batch_x, self.model)
            self.loss = self.loss_func(self.pred, self.batch_y)
        self('after_train_loss')
        # backward
        self.scaler.scale(self.loss).backward()
        self.scaler.step(self.optimizer)
        self.scaler.update()
        self.optimizer.zero_grad()
        self('after_train_batch')

    def validate_batch(self):
        self('before_valid_batch')
        # forward
        with torch.no_grad():
            with amp.autocast():
                self.pred = self.get_pred(self.batch_x, self.model)
                self.loss = self.loss_func(self.pred, self.batch_y)
        self('after_valid_loss')



from matplotlib import pyplot as plt
import os


def show_data_img(data):
    plt.figure()
    plt.imshow(data['image'])
    plt.figtext(0, 0, f"score: {data['target']}", fontsize=10)


def find_incremental_filename(path):
    i = 0
    while os.path.exists(f"{path}/learner{i}.pkl"):
        i += 1
    return f"{path}/learner{i}.pkl"


def shuffle_df(df):
    """
    shuffle a pandas dataframe
    """
    return df.sample(frac=1).reset_index(drop=True)


def show_dl(dl, attr=("shape"), first_n=1):
    counter = 0
    for x in dl:
        print("-"*20)
        if "shape" in attr:
            print({k: v.shape for k, v in x.items()})
        if "all" in attr:
            print(f"x: {x}")

        counter += 1
        if counter == first_n:
            break



import sys
import numpy as np
import torch
from fastai.basics import GetAttr, store_attr
from tqdm import tqdm
import matplotlib.pyplot as plt
import wandb
import logging
from dataclasses import asdict


class Callback(GetAttr):
    _default = 'learner'


class TestFreezing(Callback):
    def before_train_epoch(self):
        self.begin_params = []
        check_first_n = 50
        current_n = 0
        for param in self.learner.model.parameters():
            self.begin_params.append(param)
            current_n += 1
            if current_n >= check_first_n:
                break

    def after_epoch(self):
        self.after_params = []
        check_first_n = 50
        current_n = 0
        for param in self.learner.model.parameters():
            self.after_params.append(param)
            current_n += 1
            if current_n >= check_first_n:
                break


class SaveModel(Callback):
    '''
    Save model (1) after each epoch, (2) after the best validation loss
    Make sure this callback is the last one in the list of callbacks
    '''

    def after_validate(self):
        if self.learner.is_updated_best_valid_loss:
            torch.save(self.model.state_dict(),
                       f"{self.learner.config.path}/{self.model_name}_best_valid.pth")

    def after_epoch(self):
        if self.epoch > 1:  # don't save the first epoch to save memory
            torch.save(self.model.state_dict(),
                       f"{self.learner.config.path}/{self.model_name}_epoch{self.epoch}.pth")


class TrackLoss(Callback):
    '''
    We always track epoch train loss, but we validate periodically during the epoch, so the
    valid loss is named interval_valid_loss as opposed to epoch_valid_loss.
    '''

    def __init__(self):
        # store the loss for every step / batch
        self.train_losses = []
        self.valid_losses = []
        self.best_valid_loss = float("inf")

    def before_epoch(self):
        self.epoch_train_loss = []

    def after_train_loss(self):
        self.epoch_train_loss.append(float(self.loss))

    def before_validate(self):
        self.interval_valid_loss = []
        self.correct_count = 0

    def after_valid_loss(self):
        self.interval_valid_loss.append(float(self.loss))
        # get batch correct count
        pred = torch.argmax(self.learner.pred, 1)
        self.correct_count += pred.eq(self.batch_y).sum(
        ).item()

    def after_validate(self):
        # update valid loss
        self.valid_losses += self.interval_valid_loss
        self.learner.is_updated_best_valid_loss = self.update_best_valid_loss_maybe(
            np.mean(self.interval_valid_loss))
        # update accuracy
        self.acc = self.correct_count / len(self.dls.valid_dl.dataset)

    def after_epoch(self):
        self.train_losses += self.epoch_train_loss
        self.log_current_loss()

    def log_current_loss(self):
        if self.train_losses and self.valid_losses:
            print("*" * 90)
            print(f"epoch {self.epoch} done | avg epoch train loss: {np.mean(self.epoch_train_loss)} | avg current valid loss: {np.mean(self.interval_valid_loss)} | acc: {self.acc}")
            print("*" * 90)
            print()
        else:
            tqdm.write("Warning: one of the losses is None")

    def update_best_valid_loss_maybe(self, cur_model_val_loss):
        # update best valid loss if applicable
        if cur_model_val_loss < self.best_valid_loss:
            self.best_valid_loss = cur_model_val_loss
            tqdm.write(
                f">>>>> best valid loss updated: {self.best_valid_loss}\r")
            return True
        return False

    def after_fit(self):
        if self.save_learner:
            self.save()


class WandbTrackAndSave(TrackLoss):
    """
    Example: 
        ```
        WandbTrackAndSave("hello", {"learning_rate": self.lr})
        ```
    """

    def __init__(self, project, model_name):
        super().__init__()
        store_attr("project, model_name", self)

    def before_fit(self):
        self.best_valid_path = f"{self.learner.config.path}/{self.model_name}_best_valid.pth"
        self.wandb_run = wandb.init(project=self.project,
                                    config=asdict(self.learner.config))
        wandb.watch(self.model)  # track gradients

    def after_train_loss(self):
        step_train_loss = float(self.loss)
        self.epoch_train_loss.append(step_train_loss)
        wandb.log({"train_loss": step_train_loss})

    def after_valid_loss(self):
        step_valid_loss = float(self.loss)
        self.interval_valid_loss.append(step_valid_loss)
        wandb.log({"valid_loss": step_valid_loss})
        # get batch correct count (TODO: refactor & use `super().after_valid_loss`)
        pred = torch.argmax(self.learner.pred, 1)
        self.correct_count += pred.eq(self.batch_y).sum(
        ).item()

    def after_validate(self):
        super().after_validate()
        if self.learner.is_updated_best_valid_loss:
            torch.save(self.model.state_dict(),
                       self.best_valid_path)
        wandb.log({"accuracy": self.acc})

    def after_fit(self):
        if self.save_learner:
            self.save()
            wandb.save(f"{self.learner.config.path}/learner.pkl")
            wandb.save(f"{self.learner.config.path}/dls.pkl")
        self.wandb_run.finish()


class Debugger(Callback):
    def before_fit(self):
        pass


class CudaCallback(Callback):
    def __init__(self, device="cuda"):
        self.device = device
        logging.warning(f"device is: {device}")

    def before_fit(self):
        self.model.to(self.device)
        if self.get_pred:
            self.get_pred.to(self.device)

    def before_batch(self):
        self.learner.batch_x = {k: v.to(self.device)
                                for k, v in self.batch_x.items()}
        self.learner.batch_y = self.batch_y.to(self.device)

    def before_train_batch(self):
        self.before_batch()

    def before_valid_batch(self):
        self.before_batch()


class LrRecorder(Callback):
    def __init__(self, param_group_length):
        self.lrs = [[] for _ in range(param_group_length)]

    def before_train_batch(self):
        for i, param_group in enumerate(self.learner.optimizer.param_groups):
            self.lrs[i].append(param_group['lr'])

    def plot_lrs(self):
        for i in range(len(self.lrs)):
            if i == 0:
                label = 'tail group'
            elif i == len(self.lrs)-1:
                label = 'head group'
            else:
                label = f"group {i-1}"
            plt.plot(self.lrs[i], label=label)
        plt.legend(loc="upper right")


class Scheduler(Callback):
    def after_train_batch(self):
        self.scheduler.step()


class ProgressBar(Callback):
    def before_epoch(self):
        self.learner.pbar = tqdm(total=len(
            self.dls.train_dl), position=0, desc=f"epoch {self.epoch}", leave=True, file=sys.stdout)

    def after_train_batch(self):
        self.learner.pbar.update(1)

    def after_epoch(self):
        self.learner.pbar.close()



class DataLoaderGroup():
    def __init__(self, train_dl, valid_dl, test_dl=None):
        self.train_dl, self.valid_dl, self.test_dl = train_dl, valid_dl, test_dl

    @property
    def train_ds(self):
        return self.train_dl.dataset

    @property
    def valid_ds(self):
        return self.valid_dl.dataset

    @property
    def test_ds(self):
        return self.test_dl.dataset



from setuptools import setup, find_packages

with open("README.md") as f:
    long_description = f.read()


with open("requirements.txt") as f:
    requirements = [req.strip() for req in f]

setup(
    name='torchspider',
    version='0.0.12',
    description='A Pytorch-based framework',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Gary Lai',
    author_email="glai9665@gmail.com",
    packages=['torchspider'],
    package_dir={'': 'src'},
    install_requires=requirements,
    url="https://github.com/ghlai9665/octopus",
    python_requires=">=3.6",
    classifiers=["License :: OSI Approved :: Apache Software License",
                 "Programming Language :: Python :: 3"],
    license='Apache v2.0'
)
